#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from platform_setup.setup_base import wheel_for
from platform_setup.packaging_base import PlatformDirs

wheel_for(PlatformDirs.LinuxArm64)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

# Non-stdlib imports not allowed in this file, as it is imported prior to the getdeps call

import subprocess
from glob import glob
from os.path import (
    expanduser,
    dirname,
    abspath,
    join,
)

Libnames = [
    'libpdfium.dylib',
    'pdfium.dylib',
    'libpdfium.dll',
    'pdfium.dll',
    'libpdfium.so',
    'pdfium.so',
    'libpdfium',
    'pdfium',
]


HomeDir     = expanduser('~')
SourceTree  = dirname(dirname(abspath(__file__)))
DataTree    = join(SourceTree,'data')
SB_Dir      = join(SourceTree,'sourcebuild')
ModuleDir   = join(SourceTree,'src','pypdfium2')
VersionFile = join(ModuleDir,'_version.py')


class PlatformDirs:
    Darwin64     = join(DataTree,'darwin-x64')
    DarwinArm64  = join(DataTree,'darwin-arm64')
    Linux64      = join(DataTree,'linux-x64')
    LinuxArm64   = join(DataTree,'linux-arm64')
    LinuxArm32   = join(DataTree,'linux-arm32')
    Windows64    = join(DataTree,'windows-x64')
    Windows86    = join(DataTree,'windows-x86')
    WindowsArm64 = join(DataTree,'windows-arm64')
    SourceBuild  = join(DataTree,'sourcebuild')


def run_cmd(command, cwd):
    print(command)
    subprocess.run(command, cwd=cwd, shell=True)


def call_ctypesgen(platform_dir, include_dir):
    
    headers_str = ''
    sep = ''
    
    for file in sorted(glob( join(include_dir,'*.h') )):
        headers_str += sep + '"{}"'.format(file)
        sep = ' '
    
    bindings_file = join(platform_dir,'_pypdfium.py')
    
    ctypesgen_cmd = 'ctypesgen --library pdfium --strip-build-path "{}" -L . {} -o "{}"'.format(
        platform_dir,
        headers_str,
        bindings_file,
    )
    #print(ctypesgen_cmd)
    subprocess.run(
        ctypesgen_cmd,
        stdout = subprocess.PIPE,
        cwd    = platform_dir,
        shell  = True,
    )
    
    with open(bindings_file, 'r') as file_reader:
        text = file_reader.read()
        #text = text.split('\n"""\n', maxsplit=1)[1]
        text = text.replace(platform_dir, '.')
        text = text.replace(HomeDir, '~')
    
    with open(bindings_file, 'w') as file_writer:
        file_writer.write(text)


def _get_ver_namespace():
    
    ver_namespace = {}
    with open(VersionFile) as fh:
        exec(fh.read(), ver_namespace)
    
    return ver_namespace


_ver_namespace = _get_ver_namespace()

def extract_version(variable_str):
    return _ver_namespace[variable_str]



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

import ctypes
from os.path import abspath
from typing import (
    Union,
    Tuple,
    BinaryIO,
)
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.error_handler import handle_pdfium_error


def is_buffer(obj) -> bool:
    """
    Check whether an object is a byte buffer that implements ``seek()``, ``tell()``,
    and ``readinto()``.
    
    Returns:
        :data:`True` if the object implements all required methods, :data:`False` otherwise.
    """
    
    if all( callable(getattr(obj, a, None)) for a in ('seek', 'tell', 'readinto') ):
        return True
    else:
        return False


class _reader_class:
    """
    Class that implements the callback for ``FPDF_FILEACCESS.m_GetBlock()``, to incrementally
    read file data from a buffer.
    """
    
    def __init__(self, buffer):
        self.buffer = buffer
    
    def __call__(self, param, position, p_buf, size):
        self.buffer.seek(position)
        c_buf = (ctypes.c_char * size).from_address( ctypes.addressof(p_buf.contents) )
        self.buffer.readinto(c_buf)
        return 1


class LoaderData:
    """
    Class to store data associated to an ``FPDF_DOCUMENT`` that was opened using
    ``FPDF_LoadCustomDocument()``.
    
    Parameters:
        file_handle:
            File buffer that implements the ``close()`` method.
        reader_instance:
            The callable object for file access (a :class:`._reader_class` instance
            wrapped with the corresponding :func:`ctypes.CFUNCTYPE`).
    """
    
    def __init__(
            self,
            file_handle = None,
            reader_instance = None,
        ):
        self.file_handle = file_handle
        self.reader_instance = reader_instance
    
    def close(self):
        
        # access the reader variable again to make sure that even a heavily optimising
        # interpreter would not prematurely delete the object
        id(self.reader_instance)
        
        if self.file_handle is not None:
            self.file_handle.close()


def open_pdf_buffer(
        buffer: BinaryIO,
        password: Union[str, bytes] = None,
    ) -> Tuple[pdfium.FPDF_DOCUMENT, LoaderData]:
    """
    Open a PDF document incrementally from a byte buffer using ``FPDF_LoadCustomDocument()`` and
    a ctypes callback function.
    
    Parameters:
        buffer:
            A byte buffer as defined in :func:`.is_buffer`.
        password:
            A password to unlock the document, if encrypted.
    
    Returns:
        The handle to a PDFium document, and a :class:`.LoaderData` object to store associated
        file access data.
    
    See also :func:`.open_pdf_auto`. **The same warnings apply!**
    """
    
    if not is_buffer(buffer):
        raise ValueError("Buffer must implement the methods seek(), tell(), and readinto().")
    
    buffer.seek(0, 2)
    file_len = buffer.tell()
    buffer.seek(0)
    
    FuncType = ctypes.CFUNCTYPE(
        # restype
        ctypes.c_int,
        # argtypes
        ctypes.POINTER(None),
        ctypes.c_ulong,
        ctypes.POINTER(ctypes.c_ubyte),
        ctypes.c_ulong,
    )
    
    fileaccess = pdfium.FPDF_FILEACCESS()
    fileaccess.m_FileLen = file_len
    fileaccess.m_GetBlock = FuncType( _reader_class(buffer) )
    
    pdf = pdfium.FPDF_LoadCustomDocument(ctypes.byref(fileaccess), password)
    ld_data = LoaderData(buffer, fileaccess.m_GetBlock)
    
    if pdfium.FPDF_GetPageCount(pdf) < 1:
        handle_pdfium_error(False)
    
    return pdf, ld_data


def open_pdf_native(
        filepath: str,
        password: Union[str, bytes] = None,
    ) -> Tuple[pdfium.FPDF_DOCUMENT, LoaderData]:
    """    
    Open a PDF document from a file path, managing all file access natively in Python using
    :func:`.open_pdf_buffer`, without having to load the whole file into memory at once.
    This ensures independence from file access in PDFium, which currently has issues with
    widestring paths on Windows. However, it may be slower to a certain extent due to object
    conversions between C and Python involved in the callback.
    
    See also :func:`.open_pdf_auto`. **The same warnings apply!**
    """
    
    file_handle = open(abspath(filepath), 'rb')
    return open_pdf_buffer(file_handle, password)



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
from os.path import (
    join,
    basename,
)
import shutil
import sysconfig
import setuptools
from glob import glob
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

from platform_setup.packaging_base import (
    Libnames,
    SourceTree,
    ModuleDir,
    PlatformDirs,
    extract_version,
)


def _get_bdist(whl_tag):
    
    class bdist (_bdist_wheel):
        def __init__(self, *args, **kwargs):
            _bdist_wheel.__init__(self, *args, **kwargs)
            self.python_tag = 'py3'
            self.plat_name = whl_tag
            self.plat_name_supplied = True
    
    return bdist


def _clean():
    
    build_cache    = join(SourceTree,'build')
    bindings_file  = join(ModuleDir,'_pypdfium.py')
    
    libpaths = []
    for name in Libnames:
        libpaths.append( join(ModuleDir, name) )
    
    files = [bindings_file, *libpaths]
    
    if os.path.exists(build_cache):
        shutil.rmtree(build_cache)
    
    for file in files:
        if os.path.exists(file):
            os.remove(file)


def _copy_bindings(platform_dir):
    
    # non-recursively collect all objects from the platform directory
    for src_path in glob(join(platform_dir,'*')):
        
        # copy platform-specific files into the sources, excluding possible directories
        if os.path.isfile(src_path):
            dest_path = join(ModuleDir, basename(src_path))
            shutil.copy(src_path, dest_path)


def _get_tag(plat_dir):
    if plat_dir is PlatformDirs.Darwin64:
        return 'macosx_10_11_x86_64'
    elif plat_dir is PlatformDirs.DarwinArm64:
        return 'macosx_11_0_arm64'
    elif plat_dir is PlatformDirs.Linux64:
        return 'manylinux_2_17_x86_64'
    elif plat_dir is PlatformDirs.LinuxArm64:
        return 'manylinux_2_17_aarch64'
    elif plat_dir is PlatformDirs.LinuxArm32:
        return 'manylinux_2_17_armv7l'
    elif plat_dir is PlatformDirs.Windows64:
        return 'win_amd64'
    elif plat_dir is PlatformDirs.Windows86:
        return 'win32'
    elif plat_dir is PlatformDirs.WindowsArm64:
        return 'win_arm64'
    elif plat_dir is PlatformDirs.SourceBuild:
        return sysconfig.get_platform()
    else:
        raise ValueError( "Unknown platform directory {}".format(plat_dir) )

   
SetupKws = dict(
    version = extract_version('V_PYPDFIUM2'),
)

def wheel_for(platform_dir):
    _clean()
    _copy_bindings(platform_dir)
    tag = _get_tag(platform_dir)
    setuptools.setup(
        cmdclass = {'bdist_wheel': _get_bdist(tag)},
        package_data = {'': Libnames},
        **SetupKws,
    )
    _clean()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# Download the PDFium binaries and generate ctypes bindings

import os
import sys
import shutil
import tarfile
import zipfile
import argparse
import traceback
import subprocess
from os.path import (
    join,
    dirname,
    basename,
)
from urllib import request
from concurrent.futures import ThreadPoolExecutor

from platform_setup.packaging_base import (
    PlatformDirs,
    VersionFile,
    extract_version,
    call_ctypesgen,
)


ReleaseURL = "https://github.com/bblanchon/pdfium-binaries/releases/download/chromium%2F"
ReleaseExtension = "tgz"
ReleaseFiles = {
    PlatformDirs.Darwin64     : 'pdfium-mac-x64',
    PlatformDirs.DarwinArm64  : 'pdfium-mac-arm64',
    PlatformDirs.Linux64      : 'pdfium-linux-x64',
    PlatformDirs.LinuxArm64   : 'pdfium-linux-arm64',
    PlatformDirs.LinuxArm32   : 'pdfium-linux-arm',
    PlatformDirs.Windows64    : 'pdfium-win-x64',
    PlatformDirs.Windows86    : 'pdfium-win-x86',
    PlatformDirs.WindowsArm64 : 'pdfium-win-arm64',
}

def _set_versions(*versions_list):
    
    with open(VersionFile, 'r') as fh:
        content = fh.read()
    
    for variable, current_ver, new_ver in versions_list:
        
        template = "{} = {}"
        previous = template.format(variable, current_ver)
        updated = template.format(variable, new_ver)
        
        print( "'{}' -> '{}'".format(previous, updated) )
        content = content.replace(previous, updated, 1)
    
    with open(VersionFile, 'w') as fh:
        fh.write(content)


def get_latest_version():
    
    git_ls = subprocess.run(
        "git ls-remote https://github.com/bblanchon/pdfium-binaries.git",
        stdout = subprocess.PIPE,
        stderr = subprocess.STDOUT,
        shell  = True,
    )
    
    git_ls = git_ls.stdout.decode('UTF-8')
    tag = git_ls.split('\t')[-1].replace('\n', '')
    version = int(tag.split('/')[-1])
    
    return version


def handle_versions(latest_version):
    
    v_minor = extract_version("V_MINOR")
    v_libpdfium = extract_version("V_LIBPDFIUM")
    
    if v_libpdfium < latest_version:
        print("New PDFium build")
        _set_versions(
            ("V_MINOR", v_minor, v_minor+1),
            ("V_LIBPDFIUM", v_libpdfium, latest_version),
        )
        
    else:
        print("No new PDFium build - will re-create bindings without incrementing version")


def clear_data():
    for dirpath in ReleaseFiles:
        if os.path.isdir(dirpath):
            shutil.rmtree(dirpath)


def _get_package(args):
    
    dirpath, file_url, file_path = args
    print("Downloading {} -> {}".format(file_url, file_path))
    
    try:
        request.urlretrieve(file_url, file_path)
    except Exception:
        traceback.print_exc()
        return None
    
    return dirpath, file_path


def download_releases(latest_version, download_files):
    
    base_url = "{}{}/".format(ReleaseURL, latest_version)
    args = []
    
    for dirpath, arcname in download_files.items():
        
        filename = "{}.{}".format(arcname, ReleaseExtension)
        file_url = base_url + filename
        
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        
        file_path = join(dirpath, filename)
        args.append( (dirpath, file_url, file_path) )
    
    archives = {}
    
    with ThreadPoolExecutor() as pool:
        
        for output in pool.map(_get_package, args):
            
            if output is not None:
                dirpath, file_path = output
                archives[dirpath] = file_path
    
    return archives


def unpack_archives(archives):
    
    for file in archives.values():
        
        extraction_path = join(dirname(file), 'build_tar')
        
        if ReleaseExtension == 'tgz':
            arc_opener = tarfile.open
        elif ReleaseExtension == 'zip':
            arc_opener = zipfile.ZipFile
        else:
            raise ValueError("Unknown archive extension {}".format(ReleaseExtension))
        
        with arc_opener(file) as archive:
            archive.extractall(extraction_path)
        
        os.remove(file)


def generate_bindings(archives): 
    
    for platform_dir in archives.keys():
        
        build_dir = join(platform_dir,'build_tar')
        bin_dir = join(build_dir,'lib')
        dirname = basename(platform_dir)
        
        if dirname.startswith('windows'):
            target_name = 'pdfium.dll'
            bin_dir = join(build_dir,'bin')
        elif dirname.startswith('darwin'):
            target_name = 'pdfium.dylib'
        elif dirname.startswith('linux'):
            target_name = 'pdfium'
        else:
            raise ValueError("Unknown platform directory name '{}'".format(dirname))
        
        current_name = os.listdir(bin_dir)[0]
        shutil.move(join(bin_dir, current_name), join(platform_dir, target_name))
        
        call_ctypesgen(platform_dir, join(build_dir,'include'))
        shutil.rmtree(build_dir)


def parse_args(argv):
    parser = argparse.ArgumentParser(
        description = "Download pre-built PDFium packages and generate the bindings,",
    )
    parser.add_argument(
        '--platforms', '-p',
        metavar = 'P',
        nargs = '*',
    )
    return parser.parse_args(argv)


def get_download_files(args):
    
    platforms = args.platforms
    if platforms is None:
        return ReleaseFiles
    
    short_platforms = {}
    for key in ReleaseFiles.keys():
        short_platforms[basename(key)] = key
    
    download_files = {}
    
    for short_name in args.platforms:
        
        if short_name in short_platforms:
            long_name = short_platforms[short_name]
            download_files[long_name] = ReleaseFiles[long_name]
        else:
            available_keys = [k for k in short_platforms.keys()]
            raise ValueError(
                "Unknown platform name '{}'. Available keys are {}.".format(short_name, available_keys)
            )
    
    return download_files


def main(argv=sys.argv[1:]):
    
    args = parse_args(argv)
    download_files = get_download_files(args)
    
    latest_version = get_latest_version()
    handle_versions(latest_version)
    clear_data()
    
    archives = download_releases(latest_version, download_files)
    unpack_archives(archives)
    generate_bindings(archives)


if __name__ == '__main__':
    main()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import shutil
from importlib.util import find_spec
from platform_setup.packaging_base import run_cmd


SysCommands = [
    'git',
    'gcc',
]
NB_SysCommands = [
    'gn',
    'ninja',
    'clang',
    'lld',
]


def _pip_install(pkg):
    run_cmd("python3 -m pip install {}".format(pkg), cwd=None)

def _install_ctypesgen():
    if not shutil.which('ctypesgen') or not find_spec('ctypesgen'):
        _pip_install('ctypesgen')

def _install_wheel():
    if not find_spec('wheel'):
        _pip_install('wheel')


def install_pydeps():
    _install_ctypesgen()
    _install_wheel()


def check_sysdeps(sys_commands):
    
    for dep_name in sys_commands:
        
        dep_binary = shutil.which(dep_name)
        
        if dep_binary is None:
            print(
                "Missing system dependency: {} - please install it using your package manager.".format(dep_name),
                file = sys.stderr,
            )
        else:
            print("Found installation of system dependency {}".format(dep_name))


def main(prefer_st=False):
    
    sys_commands = SysCommands.copy()
    if prefer_st:
        sys_commands += NB_SysCommands
    
    check_sysdeps(sys_commands)
    install_pydeps()


if __name__ == '__main__':
    main()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# imports prior to getdeps call (non-stdlib dependencies forbidden)

import sysconfig
from os.path import join, basename
from platform_setup import getdeps
from platform_setup.packaging_base import SourceTree, PlatformDirs


StatusFile = join(SourceTree, 'platform_setup', 'setup_status.txt')

def check_presetup() -> bool:
    with open(StatusFile, 'r') as file_handle:
        content = file_handle.read().strip()
    if content == 'InitialState':
        return True
    elif content == 'PreSetupDone':
        return False
    else:
        raise ValueError( "Invalid content in setup status file: '{}'".format(content) )

def presetup_done():
    with open(StatusFile, 'w') as file_handle:
        file_handle.write("PreSetupDone")

W_Presetup = check_presetup()

if W_Presetup:
    getdeps.main()


# imports after getdeps call (non-stdlib dependencies allowed)

from platform_setup.setup_base import wheel_for
from platform_setup import build_pdfium, update_pdfium


class PlatformManager:
    
    def __init__(self):
        
        plat_name = sysconfig.get_platform().lower()
        for char in ('-', '.'):
            plat_name = plat_name.replace(char, '_')
        
        self.plat_name = plat_name
    
    def _is_platform(self, start, end):
        if self.plat_name.startswith(start):
            if self.plat_name.endswith(end):
                return True
        return False
    
    def is_darwin_arm64(self):
        return self._is_platform('macosx', 'arm64')
    def is_darwin_x64(self):
        return self._is_platform('macosx', 'x86_64')
    def is_linux_arm32(self):
        return self._is_platform('linux', 'armv7l')
    def is_linux_arm64(self):
        return self._is_platform('linux', 'aarch64')
    def is_linux_x64(self):
        return self._is_platform('linux', 'x86_64')
    def is_windows_arm64(self):
        return self._is_platform('win', 'arm64')    
    def is_windows_x64(self):
        return self._is_platform('win', 'amd64')
    def is_windows_x86(self):
        return self._is_platform('win32', '')


def _setup(platform_dir):
    if W_Presetup:
        update_pdfium.main( ['-p', basename(platform_dir)] )
    wheel_for(platform_dir)


def main():
    
    plat = PlatformManager()
    
    if plat.is_darwin_arm64():
        _setup(PlatformDirs.DarwinArm64)
    elif plat.is_darwin_x64():
        _setup(PlatformDirs.Darwin64)
    elif plat.is_linux_arm32():
        _setup(PlatformDirs.LinuxArm32)
    elif plat.is_linux_arm64():
        _setup(PlatformDirs.LinuxArm64)
    elif plat.is_linux_x64():
        _setup(PlatformDirs.Linux64)
    elif plat.is_windows_arm64():
        _setup(PlatformDirs.WindowsArm64)
    elif plat.is_windows_x64():
        _setup(PlatformDirs.Windows64)
    elif plat.is_windows_x86():
        _setup(PlatformDirs.Windows86)
    else:
        # Platform without pre-built binaries - trying a regular sourcebuild
        if W_Presetup:
            args = build_pdfium.parse_args([])
            build_pdfium.main(args)
        wheel_for(PlatformDirs.SourceBuild)
    
    presetup_done()


if __name__ == '__main__':
    main()
